@@ -3,10 +3,11 @@ import asyncio
 from dataclasses import dataclass
 from datetime import datetime
 from enum import Enum
-import logging
-from typing import Any, Callable, Coroutine, Dict, List, Optional, Set, Type, Union
+from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union
 
-from simplipy.entity import Entity, EntityTypes
+from simplipy.const import LOGGER
+from simplipy.entity import EntityTypes
+from simplipy.entity.factory import EntityFactory
 from simplipy.errors import PinError, SimplipyError
 from simplipy.lock import Lock
 from simplipy.sensor.v2 import SensorV2
@@ -14,104 +15,24 @@ from simplipy.sensor.v3 import SensorV3
 from simplipy.util.dt import utc_from_timestamp
 from simplipy.util.string import convert_to_underscore
 
-_LOGGER: logging.Logger = logging.getLogger(__name__)
+if TYPE_CHECKING:
+    from simplipy.api import API
 
 VERSION_V2 = 2
 VERSION_V3 = 3
 
 EVENT_SYSTEM_NOTIFICATION = "system_notification"
 
-CONF_DEFAULT: str = "default"
-CONF_DURESS_PIN: str = "duress"
-CONF_MASTER_PIN: str = "master"
+CONF_DEFAULT = "default"
+CONF_DURESS_PIN = "duress"
+CONF_MASTER_PIN = "master"
 
-DEFAULT_MAX_USER_PINS: int = 4
-MAX_PIN_LENGTH: int = 4
-RESERVED_PIN_LABELS: Set[str] = {CONF_DURESS_PIN, CONF_MASTER_PIN}
-
-ENTITY_MAP: Dict[int, dict] = {
-    VERSION_V2: {CONF_DEFAULT: SensorV2},
-    VERSION_V3: {CONF_DEFAULT: SensorV3, EntityTypes.lock: Lock},
-}
+DEFAULT_MAX_USER_PINS = 4
+MAX_PIN_LENGTH = 4
+RESERVED_PIN_LABELS = {CONF_DURESS_PIN, CONF_MASTER_PIN}
 
 
-def guard_from_missing_data(default_value: Any = None):
-    """Guard a missing property by returning a set value."""
-
-    def decorator(func):
-        """Decorate."""
-
-        def wrapper(system):
-            """Call the function and handle any issue."""
-            try:
-                return func(system)
-            except KeyError as err:
-                _LOGGER.warning(
-                    "SimpliSafe didn't return data for property: %s", func.__name__
-                )
-                _LOGGER.debug(err)
-                return default_value
-
-        return wrapper
-
-    return decorator
-
-
-def create_pin_payload(pins: dict, *, version: int = VERSION_V3) -> dict:
-    """Create the appropriate PIN payload for the provided version."""
-    empty_user_index: int
-    idx: int
-    label: str
-    payload: dict
-    pin: str
-
-    if version == VERSION_V2:
-        payload = {
-            "pins": {
-                CONF_DURESS_PIN: {"value": pins.pop(CONF_DURESS_PIN)},
-                "pin1": {"value": pins.pop(CONF_MASTER_PIN)},
-            }
-        }
-
-        empty_user_index = len(pins)
-        for idx, (label, pin) in enumerate(pins.items()):
-            payload["pins"][f"pin{idx + 2}"] = {"name": label, "value": pin}
-
-        for idx in range(DEFAULT_MAX_USER_PINS - empty_user_index):
-            payload["pins"][f"pin{str(idx + 2 + empty_user_index)}"] = {
-                "name": "",
-                "pin": "",
-            }
-    else:
-        payload = {
-            "pins": {
-                CONF_DURESS_PIN: {"pin": pins.pop(CONF_DURESS_PIN)},
-                CONF_MASTER_PIN: {"pin": pins.pop(CONF_MASTER_PIN)},
-                "users": {},
-            }
-        }
-
-        for idx, (label, pin) in enumerate(pins.items()):
-            payload["pins"]["users"][str(idx)] = {"name": label, "pin": pin}
-
-        empty_user_index = len(pins)
-        for idx in range(DEFAULT_MAX_USER_PINS - empty_user_index):
-            payload["pins"]["users"][str(idx + empty_user_index)] = {
-                "name": "",
-                "pin": "",
-            }
-
-    return payload
-
-
-def get_entity_class(
-    entity_type: EntityTypes, *, version: int = VERSION_V3
-) -> Type[Entity]:
-    """Return the appropriate entity class based on version and entity type."""
-    return ENTITY_MAP[version].get(entity_type, ENTITY_MAP[version][CONF_DEFAULT])
-
-
-@dataclass(frozen=True)  # pylint: disable=too-many-instance-attributes
+@dataclass(frozen=True)
 class SystemNotification:
     """Define a representation of a system notification."""
 
@@ -146,7 +67,37 @@ class SystemStates(Enum):
     unknown = 99
 
 
-class System:
+def coerce_state_from_raw_value(value: str) -> SystemStates:
+    """Return a proper state from a string input."""
+    try:
+        return SystemStates[convert_to_underscore(value)]
+    except KeyError:
+        LOGGER.error("Unknown system state: %s", value)
+        return SystemStates.unknown
+
+
+def guard_from_missing_data(default_value: Any = None):
+    """Guard a missing property by returning a set value."""
+
+    def decorator(func):
+        """Decorate."""
+
+        def wrapper(system):
+            """Call the function and handle any issue."""
+            try:
+                return func(system)
+            except KeyError:
+                LOGGER.warning(
+                    "SimpliSafe didn't return data for property: %s", func.__name__
+                )
+                return default_value
+
+        return wrapper
+
+    return decorator
+
+
+class System:  # pylint: disable=too-many-instance-attributes
     """Define a system.
 
     Note that this class shouldn't be instantiated directly; it will be instantiated as
@@ -160,22 +111,18 @@ class System:
     :type location_info: ``dict``
     """
 
-    def __init__(
-        self,
-        request: Callable[..., Coroutine],
-        get_subscription_data: Callable[..., Coroutine],
-        location_info: dict,
-    ) -> None:
+    def __init__(self, api: "API", system_id: int) -> None:
         """Initialize."""
-        self._get_subscription_data: Callable[..., Coroutine] = get_subscription_data
-        self._location_info: dict = location_info
-        self._notifications: List[
-            SystemNotification
-        ] = self._generate_system_notification_objects()
-        self._request: Callable[..., Coroutine] = request
-        self._state: SystemStates = self._coerce_state_from_raw_value(
-            location_info["system"].get("alarmState")
-        )
+        self._api = api
+        self._entity_factory = EntityFactory(api, self)
+        self._system_id = system_id
+
+        # These will get filled in after initial update:
+        self._notifications: List[SystemNotification] = []
+        self._state = SystemStates.unknown
+        self.entity_data: Dict[str, dict] = {}
+        self.system_data = api.subscription_data[system_id]
+
         self.locks: Dict[str, Lock] = {}
         self.sensors: Dict[str, Union[SensorV2, SensorV3]] = {}
 
@@ -186,7 +133,7 @@ class System:
 
         :rtype: ``str``
         """
-        return self._location_info["street1"]
+        return self.system_data["location"]["street1"]
 
     @property  # type: ignore
     @guard_from_missing_data(False)
@@ -195,7 +142,7 @@ class System:
 
         :rtype: ``bool``
         """
-        return self._location_info["system"]["isAlarming"]
+        return self.system_data["location"]["system"]["isAlarming"]
 
     @property  # type: ignore
     @guard_from_missing_data()
@@ -204,7 +151,7 @@ class System:
 
         :rtype: ``str``
         """
-        return self._location_info["system"]["connType"]
+        return self.system_data["location"]["system"]["connType"]
 
     @property
     def notifications(self) -> List[SystemNotification]:
@@ -221,7 +168,7 @@ class System:
 
         :rtype: ``str``
         """
-        return self._location_info["system"]["serial"]
+        return self.system_data["location"]["system"]["serial"]
 
     @property
     def state(self) -> SystemStates:
@@ -238,7 +185,7 @@ class System:
 
         :rtype: ``int``
         """
-        return self._location_info["sid"]
+        return self._system_id
 
     @property  # type: ignore
     @guard_from_missing_data()
@@ -247,7 +194,7 @@ class System:
 
         :rtype: ``int``
         """
-        return self._location_info["system"]["temperature"]
+        return self.system_data["location"]["system"]["temperature"]
 
     @property  # type: ignore
     @guard_from_missing_data()
@@ -256,96 +203,7 @@ class System:
 
         :rtype: ``int``
         """
-        return self._location_info["system"]["version"]
-
-    def _coerce_state_from_raw_value(self, value: Union[str, None]) -> SystemStates:
-        """Return a proper state from a string input."""
-        if not value:
-            _LOGGER.debug(
-                "SimpliSafe didn't return system state; retaining current state"
-            )
-            return self.state
-
-        try:
-            return SystemStates[convert_to_underscore(value)]
-        except KeyError:
-            _LOGGER.error("Unknown system state: %s", value)
-            return SystemStates.unknown
-
-    def _generate_system_notification_objects(self) -> List[SystemNotification]:
-        """Generate message objects from the message data stored in location_info."""
-        if self._location_info["system"].get("messages") is None:
-            _LOGGER.info("Notifications unavailable in plan")
-            return []
-
-        return [
-            SystemNotification(
-                raw_message["id"],
-                raw_message["text"],
-                raw_message["category"],
-                raw_message["code"],
-                raw_message["timestamp"],
-                link=raw_message["link"],
-                link_label=raw_message["linkLabel"],
-            )
-            for raw_message in self._location_info["system"]["messages"]
-        ]
-
-    async def _get_entities(self, cached: bool = True) -> None:
-        """Update sensors to the latest values."""
-        entities: dict = await self._get_entities_payload(cached)
-
-        _LOGGER.debug("Get entities response: %s", entities)
-
-        entity_data: dict
-        for entity_data in entities:
-            if not entity_data:
-                continue
-
-            try:
-                entity_type: EntityTypes = EntityTypes(entity_data["type"])
-            except ValueError:
-                _LOGGER.error("Unknown entity type: %s", entity_data["type"])
-                entity_type = EntityTypes.unknown
-
-            prop = self.locks if entity_type == EntityTypes.lock else self.sensors
-            if entity_data["serial"] in prop:
-                entity = prop[entity_data["serial"]]
-                entity.entity_data = entity_data
-            else:
-                klass = get_entity_class(entity_type, version=self.version)
-                prop[entity_data["serial"]] = klass(  # type: ignore
-                    self._request,
-                    self._get_entities,
-                    self.system_id,
-                    entity_type,
-                    entity_data,
-                )
-
-    async def _get_entities_payload(self, cached: bool = False) -> dict:
-        """Return the current sensor payload."""
-        raise NotImplementedError()
-
-    async def _get_settings(self, cached: bool = True) -> None:
-        """Get all system settings."""
-        pass
-
-    async def _get_system_info(self) -> None:
-        """Update information on the system."""
-        subscription_resp: dict = await self._get_subscription_data()
-        location_info: dict = next(
-            (
-                system["location"]
-                for system in subscription_resp["subscriptions"]
-                if system["sid"] == self.system_id
-            )
-        )
-
-        self._location_info = location_info
-        self._notifications = self._generate_system_notification_objects()
-        self._state = self._coerce_state_from_raw_value(
-            location_info["system"].get("alarmState")
-        )
+        return self.system_data["location"]["system"]["version"]
 
     async def _set_updated_pins(self, pins: dict) -> None:
         """Post new PINs."""
@@ -355,14 +213,49 @@ class System:
         """Raise if calling this undefined based method."""
         raise NotImplementedError()
 
+    async def _update_entity_data(self, cached: bool = True) -> None:
+        """Get all entities related to this system."""
+        await self._update_entity_data_internal(cached)
+
+        self.locks = {}
+        self.sensors = {}
+
+        for serial, entity in self.entity_data.items():
+            try:
+                entity_type = EntityTypes(entity["type"])
+            except ValueError:
+                LOGGER.error("Unknown entity type: %s", entity["type"])
+                entity_type = EntityTypes.unknown
+
+            instance = self._entity_factory.create(entity_type, serial)
+
+            if isinstance(instance, Lock):
+                self.locks[instance.serial] = instance
+            else:
+                self.sensors[instance.serial] = instance
+
+    async def _update_entity_data_internal(self, cached: bool = False) -> None:
+        """Update all entity data."""
+        raise NotImplementedError()
+
+    async def _update_settings_data(self, cached: bool = True) -> None:
+        """Update all settings data."""
+        pass
+
+    async def _update_system_data(self) -> None:
+        """Update all system data."""
+        await self._api.update_subscription_data()
+
     async def clear_notifications(self):
         """Clear all active notifications.
 
-        This will remove the notification from SimpliSafe's cloud, meaning it will no
+        This will remove the notifications from SimpliSafe's cloud, meaning they will no
         longer visible in the SimpliSafe mobile and web apps.
         """
         if self._notifications:
-            await self._request("delete", f"subscriptions/{self.system_id}/messages")
+            await self._api.request(
+                "delete", f"subscriptions/{self.system_id}/messages"
+            )
             self._notifications = []
 
     async def get_events(
@@ -378,18 +271,16 @@ class System:
         :type num_events: ``int``
         :rtype: ``list``
         """
-        params: Dict[str, Any] = {}
+        params = {}
         if from_datetime:
             params["fromTimestamp"] = round(from_datetime.timestamp())
         if num_events:
             params["numEvents"] = num_events
 
-        events_resp: dict = await self._request(
+        events_resp = await self._api.request(
             "get", f"subscriptions/{self.system_id}/events", params=params
         )
 
-        _LOGGER.debug("Events response: %s", events_resp)
-
         return events_resp.get("events", [])
 
     async def get_latest_event(self) -> dict:
@@ -397,7 +288,7 @@ class System:
 
         :rtype: ``dict``
         """
-        events: list = await self.get_events(num_events=1)
+        events = await self.get_events(num_events=1)
 
         try:
             return events[0]
@@ -425,15 +316,13 @@ class System:
         # Because SimpliSafe's API works by sending the entire payload of PINs, we
         # can't reasonably check a local cache for up-to-date PIN data; so, we fetch the
         # latest each time:
-        latest_pins: Dict[str, str] = await self.get_pins(cached=False)
+        latest_pins = await self.get_pins(cached=False)
 
         if pin_or_label in RESERVED_PIN_LABELS:
             raise PinError(f"Refusing to delete reserved PIN: {pin_or_label}")
 
         try:
-            label: str = next(
-                (k for k, v in latest_pins.items() if pin_or_label in (k, v))
-            )
+            label = next((k for k, v in latest_pins.items() if pin_or_label in (k, v)))
         except StopIteration:
             raise PinError(f"Cannot delete nonexistent PIN: {pin_or_label}")
 
@@ -472,12 +361,12 @@ class System:
         # Because SimpliSafe's API works by sending the entire payload of PINs, we
         # can't reasonably check a local cache for up-to-date PIN data; so, we fetch the
         # latest each time.
-        latest_pins: Dict[str, str] = await self.get_pins(cached=False)
+        latest_pins = await self.get_pins(cached=False)
 
         if pin in latest_pins.values():
             raise PinError(f"Refusing to create duplicate PIN: {pin}")
 
-        max_pins: int = DEFAULT_MAX_USER_PINS + len(RESERVED_PIN_LABELS)
+        max_pins = DEFAULT_MAX_USER_PINS + len(RESERVED_PIN_LABELS)
         if len(latest_pins) == max_pins and label not in RESERVED_PIN_LABELS:
             raise PinError(f"Refusing to create more than {max_pins} user PINs")
 
@@ -502,21 +391,40 @@ class System:
         :type include_system: ``bool``
         :param include_settings: Whether system settings (like PINs) should be updated
         :type include_settings: ``bool``
-        :param include_entities: Whether sensors/locks/etc. should be updated
+        :param include_entities: whether sensors/locks/etc. should be updated
         :type include_entities: ``bool``
         :param cached: Whether to used cached data.
         :type cached: ``bool``
         """
-        tasks: List[Coroutine] = []
+        tasks = []
 
         if include_system:
-            tasks.append(self._get_system_info())
+            tasks.append(self._update_system_data())
         if include_settings:
-            tasks.append(self._get_settings(cached))
+            tasks.append(self._update_settings_data(cached))
 
         await asyncio.gather(*tasks)
 
         # We await entity updates after the task pool since including it can cause
         # HTTP 409s if that update occurs out of sequence:
         if include_entities:
-            await self._get_entities(cached)
+            await self._update_entity_data(cached)
+
+        self._notifications = [
+            SystemNotification(
+                raw_message["id"],
+                raw_message["text"],
+                raw_message["category"],
+                raw_message["code"],
+                raw_message["timestamp"],
+                link=raw_message["link"],
+                link_label=raw_message["linkLabel"],
+            )
+            for raw_message in self.system_data["location"]["system"].get(
+                "messages", []
+            )
+        ]
+
+        self._state = coerce_state_from_raw_value(
+            self.system_data["location"]["system"].get("alarmState")
+        )
