@@ -1,19 +1,21 @@
 """Define a V3 (new) SimpliSafe system."""
-from enum import Enum
-import logging
-from typing import Dict
+from typing import TYPE_CHECKING, Dict
 
 import voluptuous as vol
 
+from simplipy.camera import Camera
 from simplipy.system import (
     CONF_DURESS_PIN,
     CONF_MASTER_PIN,
+    DEFAULT_MAX_USER_PINS,
     System,
-    create_pin_payload,
+    SystemStates,
+    coerce_state_from_raw_value,
     guard_from_missing_data,
 )
 
-_LOGGER: logging.Logger = logging.getLogger(__name__)
+if TYPE_CHECKING:
+    from simplipy.api import API
 
 CONF_ALARM_DURATION = "alarm_duration"
 CONF_ALARM_VOLUME = "alarm_volume"
@@ -70,13 +72,47 @@ SYSTEM_PROPERTIES_PAYLOAD_SCHEMA = vol.Schema(
 )
 
 
+def create_pin_payload(pins: dict) -> Dict[str, Dict[str, Dict[str, str]]]:
+    """Create the request payload to send for updating PINs."""
+    duress_pin = pins.pop(CONF_DURESS_PIN)
+    master_pin = pins.pop(CONF_MASTER_PIN)
+
+    payload = {
+        "pins": {
+            CONF_DURESS_PIN: {"pin": duress_pin},
+            CONF_MASTER_PIN: {"pin": master_pin},
+        }
+    }
+
+    user_pins = {}
+    for idx, (label, pin) in enumerate(pins.items()):
+        user_pins[str(idx)] = {"name": label, "pin": pin}
+
+    empty_user_index = len(pins)
+    for idx in range(DEFAULT_MAX_USER_PINS - empty_user_index):
+        user_pins[str(idx + empty_user_index)] = {
+            "name": "",
+            "pin": "",
+        }
+
+    payload["users"] = user_pins
+
+    return payload
+
+
 class SystemV3(System):
     """Define a V3 (new) system."""
 
-    def __init__(self, request, get_subscription_data, location_info) -> None:
+    def __init__(self, api: "API", system_id: int) -> None:
         """Initialize."""
-        super().__init__(request, get_subscription_data, location_info)
-        self.settings_info: dict = {}
+        super().__init__(api, system_id)
+
+        # This will be filled in by the appropriate data update methods:
+        self.camera_data: Dict[str, dict] = {}
+        self.settings_data: Dict[str, dict] = {}
+
+        self.cameras: Dict[str, Camera] = {}
+        self._update_camera_data()
 
     @property  # type: ignore
     @guard_from_missing_data()
@@ -85,7 +121,7 @@ class SystemV3(System):
 
         :rtype: ``int``
         """
-        return self.settings_info["settings"]["normal"][
+        return self.settings_data["settings"]["normal"][
             SYSTEM_PROPERTIES_VALUE_MAP["alarm_duration"]
         ]
 
@@ -97,7 +133,7 @@ class SystemV3(System):
         :rtype: ``int``
         """
         return int(
-            self.settings_info["settings"]["normal"][
+            self.settings_data["settings"]["normal"][
                 SYSTEM_PROPERTIES_VALUE_MAP["alarm_volume"]
             ]
         )
@@ -109,7 +145,7 @@ class SystemV3(System):
 
         :rtype: ``int``
         """
-        return self.settings_info["basestationStatus"]["backupBattery"]
+        return self.settings_data["basestationStatus"]["backupBattery"]
 
     @property  # type: ignore
     @guard_from_missing_data()
@@ -119,7 +155,7 @@ class SystemV3(System):
         :rtype: ``int``
         """
         return int(
-            self.settings_info["settings"]["normal"][
+            self.settings_data["settings"]["normal"][
                 SYSTEM_PROPERTIES_VALUE_MAP["chime_volume"]
             ]
         )
@@ -131,7 +167,7 @@ class SystemV3(System):
 
         :rtype: ``int``
         """
-        return self.settings_info["settings"]["normal"][
+        return self.settings_data["settings"]["normal"][
             SYSTEM_PROPERTIES_VALUE_MAP["entry_delay_away"]
         ]
 
@@ -142,7 +178,7 @@ class SystemV3(System):
 
         :rtype: ``int``
         """
-        return self.settings_info["settings"]["normal"][
+        return self.settings_data["settings"]["normal"][
             SYSTEM_PROPERTIES_VALUE_MAP["entry_delay_home"]
         ]
 
@@ -153,7 +189,7 @@ class SystemV3(System):
 
         :rtype: ``int``
         """
-        return self.settings_info["settings"]["normal"][
+        return self.settings_data["settings"]["normal"][
             SYSTEM_PROPERTIES_VALUE_MAP["exit_delay_away"]
         ]
 
@@ -164,7 +200,7 @@ class SystemV3(System):
 
         :rtype: ``int``
         """
-        return self.settings_info["settings"]["normal"][
+        return self.settings_data["settings"]["normal"][
             SYSTEM_PROPERTIES_VALUE_MAP["exit_delay_home"]
         ]
 
@@ -175,7 +211,7 @@ class SystemV3(System):
 
         :rtype: ``int``
         """
-        return self.settings_info["basestationStatus"]["gsmRssi"]
+        return self.settings_data["basestationStatus"]["gsmRssi"]
 
     @property  # type: ignore
     @guard_from_missing_data()
@@ -184,7 +220,7 @@ class SystemV3(System):
 
         :rtype: ``bool``
         """
-        return self.settings_info["settings"]["normal"][
+        return self.settings_data["settings"]["normal"][
             SYSTEM_PROPERTIES_VALUE_MAP["light"]
         ]
 
@@ -195,7 +231,7 @@ class SystemV3(System):
 
         :rtype: ``bool``
         """
-        return self._location_info["system"]["isOffline"]
+        return self.system_data["location"]["system"]["isOffline"]
 
     @property  # type: ignore
     @guard_from_missing_data(False)
@@ -204,7 +240,7 @@ class SystemV3(System):
 
         :rtype: ``bool``
         """
-        return self._location_info["system"]["powerOutage"]
+        return self.system_data["location"]["system"]["powerOutage"]
 
     @property  # type: ignore
     @guard_from_missing_data(False)
@@ -213,7 +249,7 @@ class SystemV3(System):
 
         :rtype: ``bool``
         """
-        return self.settings_info["basestationStatus"]["rfJamming"]
+        return self.settings_data["basestationStatus"]["rfJamming"]
 
     @property  # type: ignore
     @guard_from_missing_data()
@@ -222,7 +258,7 @@ class SystemV3(System):
 
         :rtype: ``int``
         """
-        return self.settings_info["settings"]["normal"][
+        return self.settings_data["settings"]["normal"][
             SYSTEM_PROPERTIES_VALUE_MAP["voice_prompt_volume"]
         ]
 
@@ -233,7 +269,7 @@ class SystemV3(System):
 
         :rtype: ``int``
         """
-        return self.settings_info["basestationStatus"]["wallPower"]
+        return self.settings_data["basestationStatus"]["wallPower"]
 
     @property  # type: ignore
     @guard_from_missing_data()
@@ -242,7 +278,7 @@ class SystemV3(System):
 
         :rtype: ``str``
         """
-        return self.settings_info["settings"]["normal"]["wifiSSID"]
+        return self.settings_data["settings"]["normal"]["wifiSSID"]
 
     @property  # type: ignore
     @guard_from_missing_data()
@@ -251,46 +287,58 @@ class SystemV3(System):
 
         :rtype: ``int``
         """
-        return self.settings_info["basestationStatus"]["wifiRssi"]
+        return self.settings_data["basestationStatus"]["wifiRssi"]
 
-    async def _get_entities_payload(self, cached: bool = True) -> dict:
+    async def _set_state(self, value: SystemStates) -> None:
+        """Set the state of the system."""
+        state_resp = await self._api.request(
+            "post", f"ss3/subscriptions/{self.system_id}/state/{value.name}"
+        )
+
+        self._state = coerce_state_from_raw_value(state_resp["state"])
+
+    async def _set_updated_pins(self, pins: dict) -> None:
+        """Post new PINs."""
+        self.settings_data = await self._api.request(
+            "post",
+            f"ss3/subscriptions/{self.system_id}/settings/pins",
+            json=create_pin_payload(pins),
+        )
+
+    def _update_camera_data(self) -> None:
+        """Update all system data."""
+        self.cameras = {}
+        for camera in self.system_data["location"]["system"].get("cameras", []):
+            uuid = camera["uuid"]
+            self.camera_data[uuid] = camera
+            self.cameras[uuid] = Camera(self, uuid)
+
+    async def _update_entity_data_internal(self, cached: bool = True) -> None:
         """Update sensors to the latest values."""
-        sensor_resp = await self._request(
+        sensor_resp = await self._api.request(
             "get",
             f"ss3/subscriptions/{self.system_id}/sensors",
             params={"forceUpdate": str(not cached).lower()},
         )
+        self.entity_data = {
+            entity["serial"]: entity for entity in sensor_resp.get("sensors", [])
+        }
 
-        return sensor_resp.get("sensors", [])
-
-    async def _get_settings(self, cached: bool = True) -> None:
+    async def _update_settings_data(self, cached: bool = True) -> None:
         """Get all system settings."""
-        settings_resp: dict = await self._request(
+        settings_resp = await self._api.request(
             "get",
             f"ss3/subscriptions/{self.system_id}/settings/normal",
             params={"forceUpdate": str(not cached).lower()},
         )
 
         if settings_resp:
-            self.settings_info = settings_resp
+            self.settings_data = settings_resp
 
-    async def _set_state(self, value: Enum) -> None:
-        """Set the state of the system."""
-        state_resp: dict = await self._request(
-            "post", f"ss3/subscriptions/{self.system_id}/state/{value.name}"
-        )
-
-        _LOGGER.debug('Set "%s" response: %s', value.name, state_resp)
-
-        self._state = self._coerce_state_from_string(state_resp["state"])
-
-    async def _set_updated_pins(self, pins: dict) -> None:
-        """Post new PINs."""
-        self.settings_info = await self._request(
-            "post",
-            f"ss3/subscriptions/{self.system_id}/settings/pins",
-            json=create_pin_payload(pins),
-        )
+    async def _update_system_data(self) -> None:
+        """Update all system data."""
+        await super()._update_system_data()
+        self._update_camera_data()
 
     async def get_pins(self, cached: bool = True) -> Dict[str, str]:
         """Return all of the set PINs, including master and duress.
@@ -302,16 +350,15 @@ class SystemV3(System):
         :type cached: ``bool``
         :rtype: ``Dict[str, str]``
         """
-        await self._get_settings(cached)
+        await self._update_settings_data(cached)
 
-        pins: Dict[str, str] = {
-            CONF_MASTER_PIN: self.settings_info["settings"]["pins"]["master"]["pin"],
-            CONF_DURESS_PIN: self.settings_info["settings"]["pins"]["duress"]["pin"],
+        pins = {
+            CONF_MASTER_PIN: self.settings_data["settings"]["pins"]["master"]["pin"],
+            CONF_DURESS_PIN: self.settings_data["settings"]["pins"]["duress"]["pin"],
         }
 
-        user_pin: dict
         for user_pin in [
-            p for p in self.settings_info["settings"]["pins"]["users"] if p["pin"]
+            p for p in self.settings_data["settings"]["pins"]["users"] if p["pin"]
         ]:
             pins[user_pin["name"]] = user_pin["pin"]
 
@@ -341,7 +388,7 @@ class SystemV3(System):
                 f"Using invalid values for system properties ({properties}): {err}"
             ) from None
 
-        settings_resp = await self._request(
+        settings_resp = await self._api.request(
             "post",
             f"ss3/subscriptions/{self.system_id}/settings/normal",
             json={
@@ -353,4 +400,4 @@ class SystemV3(System):
         )
 
         if settings_resp:
-            self.settings_info = settings_resp
+            self.settings_data = settings_resp
